package dev.fyloz.colorrecipesexplorer.logic.account

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.config.annotations.LogicComponent
import dev.fyloz.colorrecipesexplorer.config.security.filters.blacklistedJwtTokens
import dev.fyloz.colorrecipesexplorer.dtos.account.GroupDto
import dev.fyloz.colorrecipesexplorer.dtos.account.UserDto
import dev.fyloz.colorrecipesexplorer.dtos.account.UserSaveDto
import dev.fyloz.colorrecipesexplorer.dtos.account.UserUpdateDto
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.logic.BaseLogic
import dev.fyloz.colorrecipesexplorer.logic.Logic
import dev.fyloz.colorrecipesexplorer.model.account.Permission
import dev.fyloz.colorrecipesexplorer.service.account.UserService
import org.springframework.context.annotation.Lazy
import org.springframework.security.crypto.password.PasswordEncoder
import org.springframework.web.util.WebUtils
import java.time.LocalDateTime
import javax.servlet.http.HttpServletRequest

interface UserLogic : Logic<UserDto, UserService> {
    /** Gets all users which have the given [group]. */
    fun getAllByGroup(group: GroupDto): Collection<UserDto>

    /** Gets the user with the given [id]. */
    fun getById(id: Long, isSystemUser: Boolean): UserDto

    /** Saves the given [dto]. */
    fun save(dto: UserSaveDto): UserDto

    /** Updates the given [dto]. */
    fun update(dto: UserUpdateDto): UserDto

    /** Updates the last login time of the user with the given [id]. */
    fun updateLastLoginTime(id: Long, time: LocalDateTime = LocalDateTime.now()): UserDto

    /** Updates the password of the user with the given [id]. */
    fun updatePassword(id: Long, password: String): UserDto

    /** Adds the given [permission] to the user with the given [id]. */
    fun addPermission(id: Long, permission: Permission): UserDto

    /** Removes the given [permission] from the user with the given [id]. */
    fun removePermission(id: Long, permission: Permission): UserDto

    /** Logout a user. Add the authorization token of the given [request] to the blacklisted tokens. */
    fun logout(request: HttpServletRequest)
}

@LogicComponent
class DefaultUserLogic(
    service: UserService, @Lazy private val groupLogic: GroupLogic, @Lazy private val passwordEncoder: PasswordEncoder
) : BaseLogic<UserDto, UserService>(service, Constants.ModelNames.USER), UserLogic {
    override fun getAll() = service.getAll(false)

    override fun getAllByGroup(group: GroupDto) = service.getAllByGroup(group)

    override fun getById(id: Long) = getById(id, false)
    override fun getById(id: Long, isSystemUser: Boolean) =
        service.getById(id, isSystemUser) ?: throw notFoundException(value = id)

    override fun save(dto: UserSaveDto) = save(
        UserDto(
            id = dto.id,
            firstName = dto.firstName,
            lastName = dto.lastName,
            password = passwordEncoder.encode(dto.password),
            group = if (dto.groupId != null) groupLogic.getById(dto.groupId) else null,
            permissions = dto.permissions,
            isSystemUser = dto.isSystemUser
        )
    )

    override fun save(dto: UserDto): UserDto {
        throwIfIdAlreadyExists(dto.id)
        throwIfFirstNameAndLastNameAlreadyExists(dto.firstName, dto.lastName)

        return super.save(dto)
    }

    override fun update(dto: UserUpdateDto): UserDto {
        val user = getById(dto.id)

        return update(
            user.copy(
                firstName = dto.firstName,
                lastName = dto.lastName,
                group = if (dto.groupId != null) groupLogic.getById(dto.groupId) else null,
                permissions = dto.permissions
            )
        )
    }

    override fun update(dto: UserDto): UserDto {
        throwIfFirstNameAndLastNameAlreadyExists(dto.firstName, dto.lastName, dto.id)

        return super.update(dto)
    }

    override fun updateLastLoginTime(id: Long, time: LocalDateTime) = with(getById(id, true)) {
        update(this.copy(lastLoginTime = time))
    }

    override fun updatePassword(id: Long, password: String) = with(getById(id)) {
        update(this.copy(password = passwordEncoder.encode(password)))
    }

    override fun addPermission(id: Long, permission: Permission) = with(getById(id)) {
        update(this.copy(permissions = this.permissions + permission))
    }

    override fun removePermission(id: Long, permission: Permission) = with(getById(id)) {
        update(this.copy(permissions = this.permissions - permission))
    }

    override fun logout(request: HttpServletRequest) {
        val authorizationCookie = WebUtils.getCookie(request, Constants.HeaderNames.AUTHORIZATION)
        if (authorizationCookie != null) {
            val authorizationToken = authorizationCookie.value
            if (authorizationToken != null && authorizationToken.startsWith("Bearer")) {
                blacklistedJwtTokens.add(authorizationToken)
            }
        }
    }

    private fun throwIfIdAlreadyExists(id: Long) {
        if (service.existsById(id)) {
            throw alreadyExistsException(identifierName = ID_IDENTIFIER_NAME, value = id)
        }
    }

    private fun throwIfFirstNameAndLastNameAlreadyExists(firstName: String, lastName: String, id: Long? = null) {
        if (service.existsByFirstNameAndLastName(firstName, lastName, id)) {
            throw AlreadyExistsException(
                typeNameLowerCase,
                "$typeName already exists",
                "A $typeNameLowerCase with the name '$firstName $lastName' already exists",
                "$firstName $lastName",
                "fullName"
            )
        }
    }
}
